import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import busybee
import busybee._busybee as _busybee
import busybee._string_helpers as _string_helpers



import time
import unittest

from .context import busybee


#
# Test cases
#

class CorrectnessTestSuite(unittest.TestCase):

    # map(...)

    def test_map_WHEN_empty_list_THEN_empty_list(self):
        actual = busybee.map(func_add_one, [], stdout=NullStdout())
        self.assertListEqual(actual, [])

    def test_map_WHEN_executed_one_core_THEN_in_order_and_applied(self):
        actual = busybee.map(
            func=func_add_one,
            data=list(range(0, 1000)),
            processes=1,
            stdout=NullStdout(),
        )
        self.assertListEqual(actual, list(range(1, 1001)))

    def test_map_WHEN_executed_many_processes_THEN_in_order_and_applied(self):
        actual = busybee.map(
            func=func_add_one,
            data=list(range(0, 1000)),
            processes=8,
            stdout=NullStdout(),
        )
        self.assertListEqual(actual, list(range(1, 1001)))

    def test_map_WHEN_executed_all_processes_THEN_in_order_and_applied(self):
        actual = busybee.map(
            func=func_add_one,
            data=list(range(0, 1000)),
            processes='n',
            stdout=NullStdout(),
        )
        self.assertListEqual(actual, list(range(1, 1001)))

    def test_filter_WHEN_empty_list_THEN_empty_list(self):
        actual = busybee.filter(func_add_one, [], stdout=NullStdout())
        self.assertListEqual(actual, [])

    # filter(...)

    def test_filter_WHEN_executed_one_core_THEN_in_order_and_applied(self):
        actual = busybee.filter(
            func=func_is_even,
            data=list(range(0, 1000)),
            processes=1,
            stdout=NullStdout(),
        )
        self.assertListEqual(actual, list(range(0, 1000, 2)))

    def test_filter_WHEN_executed_many_processes_THEN_in_order_and_applied(self):
        actual = busybee.filter(
            func=func_is_even,
            data=list(range(0, 1000)),
            processes=8,
            stdout=NullStdout(),
        )
        self.assertListEqual(actual, list(range(0, 1000, 2)))

    def test_filter_WHEN_executed_all_processes_THEN_in_order_and_applied(self):
        actual = busybee.filter(
            func=func_is_even,
            data=list(range(0, 1000)),
            processes='n',
            stdout=NullStdout(),
        )
        self.assertListEqual(actual, list(range(0, 1000, 2)))

    # mk_dict(...)

    def test_mk_dict_WHEN_empty_THEN_returns_empty_dict(self):
        actual = busybee.mk_dict(
            func=func_add_one,
            keys=[],
            stdout=NullStdout(),
        )
        self.assertDictEqual(actual, {})

    def test_mk_dict_WHEN_one_process_THEN_in_order(self):
        actual = busybee.mk_dict(
            func=func_add_one,
            keys=[1, 2, 3, 4, 5],
            stdout=NullStdout(),
            processes=1,
        )
        self.assertDictEqual(actual, {1: 2, 2: 3, 3: 4, 4: 5, 5: 6})

    def test_mk_dict_WHEN_many_processes_THEN_in_order(self):
        actual = busybee.mk_dict(
            func=func_add_one,
            keys=[1, 2, 3, 4, 5],
            stdout=NullStdout(),
            processes=8,
        )
        self.assertDictEqual(actual, {1: 2, 2: 3, 3: 4, 4: 5, 5: 6})


class OutputTestSuite(unittest.TestCase):

    def test_map_WHEN_empty_list_THEN_warning_output(self):
        recorder = RecordingStdout()
        busybee.map(None, [], stdout=recorder)

        self.assertIn("skipping because of empty input", recorder.output)

    def test_map_WHEN_quiet_on_THEN_no_output(self):
        recorder = RecordingStdout()
        busybee.map(
            func=func_add_one,
            data=[1, 2, 3],
            quiet=True,
            stdout=recorder,
        )

        self.assertEqual("", recorder.output)

    def test_map_WHEN_executing_THEN_outputs_core_information(self):
        recorder = RecordingStdout()

        # Note: func_add_one_slow takes ~10ms. Therefore this finishes in
        # no less than 0.5s with processes=2.
        busybee.map(
            func=func_add_one_slow,
            data=list(range(0, 100)),
            processes='2',
            stdout=recorder,
        )

        self.assertIn("Start processing 100 items", recorder.output)
        self.assertIn("2 processes", recorder.output)

        self.assertIn("50/100, 50.0%", recorder.output)
        self.assertIn("avg: 5ms,", recorder.output)

        self.assertIn("Finished processing 100 items", recorder.output)
        self.assertIn("(avg: 5ms)", recorder.output)


#
# Helpers
#


class RecordingStdout():
    """Matches the `write` method of sys.stdout and appends all
    data to an internal `output` string.
    """

    def __init__(self):
        self.output = ""

    def write(self, string):
        self.output += string


class NullStdout():
    """Matches the `write` method of sys.stdout and ignores all calls."""

    def write(self, _):
        pass


def func_add_one(x):
    """Returns x + 1."""
    return x + 1


def func_add_one_slow(x):
    """Waits 10ms and returns x + 1."""
    time.sleep(10 / 1000)
    return x + 1


def func_is_even(x):
    """Returns `True` iff x is divisible by 2."""
    return x % 2 == 0



"""This file provides methods to generate the strings that are output
during the BusyBee execution."""

import math
import time


def _relative_time_string(time_seconds, no_ms=False):
    """Converts the given `time_seconds` into a relative, human-readable
    string representation.

    Args:
        time_seconds (int or float): The relative time to convert
        no_ms (bool): If set the `999ms` style representation is not chosen

    Returns:
        String: The string representation is chosen on the value of
                seconds to be either of the form `999ms`, `99.9s`,
                `59:59m`, or `99:59:59h`. If `None` is supplied, `-`
                is returned.
    """
    if time_seconds == None:
        return "-"

    milliseconds = int(time_seconds * 1000.0)
    if abs(milliseconds) < 1000 and not no_ms:
        return "%dms" % milliseconds

    seconds = time_seconds
    if abs(seconds) < 100.0:
        return "%.1fs" % (seconds)

    minutes = int(seconds / 60)
    seconds = abs(seconds - minutes * 60)
    if abs(minutes) < 60:
        return "%d:%02dm" % (minutes, seconds)

    hours = int(minutes / 60)
    minutes = abs(minutes - hours * 60)
    return "%d:%02d:%02dh" % (hours, minutes, seconds)


def _start_string(num_total, tag, num_processes):
    """Returns a string to be displayed before processing begins. It contains
    the number of total items and the number of processes. It is prefixed by the `tag`.

    Where information are not available or a division by zero would occur, a `-` or `0ms` is returned for
    that field.
    """
    fmt_string = "{tag}: Start processing {num_total} items with {num_processes} processes..."
    return fmt_string.format(
        tag=tag,
        num_total=num_total,
        num_processes=num_processes,
    )


def _finish_string(time_start, num_total, tag, current_time=lambda: time.time()):
    """Returns a string to be displayed after processing finished. It contains
    the number of processed items, the total time, and the average time per item.
    It is prefixed by the `tag`.

    Where information are not available or a division by zero would occur, a `-` or `0ms` is returned for
    that field.
    """
    current_time = current_time()
    time_delta = current_time - time_start
    time_avg = time_delta / num_total if num_total > 0 else None

    fmt_string = "{tag}: Finished processing {num_total} items in {time_delta} (avg: {time_avg})"
    return fmt_string.format(
        tag=tag,
        num_total=num_total,
        time_delta=_relative_time_string(time_delta, no_ms=True),
        time_avg=_relative_time_string(time_avg),
    )


def _progress_string(time_start, num_processed, num_total, tag, current_time=lambda: time.time()):
    """Returns a string that reflects the current progress during execution. It contains
    the number of processed items, the total number of items, progress in percent, the
    average time per item so far, and an estimate of the remaining time. It is prefixed by the `tag`.

    Where information are not available or a division by zero would occur, a `-` or `0ms` is returned for
    that field.
    """
    digits = math.log10(max(1, num_total)) + 1

    current_time = current_time()
    time_delta = current_time - time_start
    time_avg = time_delta / num_processed if num_processed > 0 else None

    percent = 100.0 * num_processed / num_total if num_total > 0 else 0.0
    items_remaining = num_total - num_processed
    time_remaining = time_avg * items_remaining if time_avg else None

    fmt_tag = "{tag}"
    fmt_items = "{num_processed: >%d}/{num_total: >%d}, {percent:4.1f}%%" % (
        digits, digits)
    fmt_times = "avg: {time_avg}, rem: {time_remaining}"
    fmt_string = "%s: %s (%s)" % (fmt_tag, fmt_items, fmt_times)

    return fmt_string.format(
        tag=tag,
        num_processed=num_processed,
        num_total=num_total,
        percent=percent,
        time_avg=_relative_time_string(time_avg),
        time_remaining=_relative_time_string(time_remaining, no_ms=True),
    )



"""The internal implementation of the busybee module. The `_map` method is
exported on module level through __init__.py."""

import math
import multiprocessing as mp
import time
import os
import sys

from ._string_helpers import _start_string, _progress_string, _finish_string

__VALUE_ERROR_INVALID_CORE_SPEC = ValueError(
    "Invalid core_spec! Try: `1`, `8`, `n/2`, `n-1`")

__VALUE_ERROR_INVALID_REL_CORE_SPEC = ValueError(
    "Invalid relative core_spec! Try: `n`, `n/2`, `n-1`")


def _parse_core_spec(core_spec, core_count=lambda: mp.cpu_count()):
    """Parses a `core_spec` that expresses the number of intended processes for execution
    of one of the high-level API calls. The core specification can be of two types.

    Type I is explicit by providing the number of processes to use.

    Type II is implicit by providing a value relative to the number of logical CPUs `n`.
    For Type II both simple division and substraction are allowed in the forms `n/c`
    and `n-c` where `c` is an integer. The result will increased if it is below 1.

    Examples for both types: `1`, `8`, `n-1`, `n/2`, `n/4`.

    Args:
        core_spec (String): The core specification to parse
        core_count (lambda: () -> int): A function to return the actual number of
                                        logical CPUs. Defaults to the number returned
                                        by the multiprocessing package

    Raises:
        ValueError: When given an invalid core spec

    Returns:
        int: The number of processes to use. Guaranteed to be >= 1
    """
    if not core_spec:
        raise __VALUE_ERROR_INVALID_CORE_SPEC

    # Type II: `core_spec` is a string with a simple formula
    if isinstance(core_spec, str) and len(core_spec) >= 1 and core_spec[0] == 'n':
        n = core_count()

        if len(core_spec) == 1:
            return n

        operator = core_spec[1]
        if len(core_spec) < 3 or operator not in "-/":
            raise __VALUE_ERROR_INVALID_REL_CORE_SPEC

        try:
            operand = int(core_spec[2:])
        except ValueError:
            raise __VALUE_ERROR_INVALID_REL_CORE_SPEC

        if operator == '-':
            return max(n - operand, 1)
        elif operator == '/':
            return max(int(n / operand), 1)
        else:  # pragma: no cover
            raise AssertionError()

    # Type I: `core_spec` is a number
    try:
        n = int(core_spec)
    except:
        raise __VALUE_ERROR_INVALID_CORE_SPEC

    if n <= 0:
        raise ValueError("If processes is a number it must be positive")

    return n


class _ProgressUpdateLimit():
    """Evaluates the progress state and indicates whether a progress update should
    be provided to the user.
    """

    def __init__(self, time_start, num_total, every_n_seconds=5, every_n_percent=50):
        self.time_start = time_start
        self.num_total = num_total

        self.time_last_update = time_start
        self.every_n_seconds = every_n_seconds

        self.percent_last_update = 0.0
        self.every_n_percent = every_n_percent

    def should_print(self, num_processed, current_time=None):
        """Returns `True` if an output should be provided to the user. Updates
        the internal state to track the most recent output.
        """
        if num_processed == self.num_total:
            # Do not print the 100% update as the finish message will follow directly afterwards
            return False

        if self.every_n_seconds:
            current_time = current_time() if current_time else time.time()
            if current_time - self.time_last_update >= self.every_n_seconds:
                self.time_last_update = current_time
                return True

        if self.every_n_percent:
            current_percent = 100.0 * num_processed / self.num_total
            if current_percent - self.percent_last_update >= self.every_n_percent:
                self.percent_last_update = current_percent
                return True

        return False


def _meta_func(args):
    """Takes args in the form `(func, data)` and calls `func(data)`."""
    func, data = args
    return func(data)


def _map(
    func,
    data,
    quiet=False,
    processes="n",
    tag="BusyBee",
    stdout=sys.stdout,
    update_every_n_seconds=5,
    update_every_n_percent=50,
):
    """Applies the given `func` to every item in `data` using up to the number of processes
    specified by `processes`. Interactive updates are provided via `stdout` following the limits
    provided by the `update_every_n_*` arguments.

    Args:
        func: The function that will be applied to the `data` items. It needs to be pickleable and
            therefore it must not be a lambda expression.

        data (list): The data that is processed by `func`. This must provide random access and `len()` support.
                     Ideally this is a simple list.

        quiet (bool): If `True`, no output is generated

        processes (int or string): A number of processes to use (e.g. 1, 8) or a simple formula expressing
                                the number of processes relative to the number of logical cores (e.g. `n`,
                                `n/2`, `n-1). Only simple substraction and division are supported.

        tag (string): A tag to be prefixed to the output. Helpful when chaining `map` operations.

        stdout: An object providing a `write` method as in `sys.stdout`.

        update_every_n_seconds (int or float): Produce a progress update every `n` seconds but not faster than
                                            the actual processing rate. Set to `None` to deactivate.

        update_every_n_percent (int or float): Produce a progress update every `n` percent of completion but
                                            not faster than the actual processing rate. Set to `None` to
                                            deactivate.

    Raises:
        ValueError: If an invalid specification is provided to the `processes` argument

    Returns:
        The processed list with items following the order of the original list.
    """
    # internal wrapper for output
    def println(string):
        if quiet:
            return
        stdout.write(string + os.linesep)

    # do not even try anything when having an empty input
    if len(data) == 0:
        println("%s: skipping because of empty input" % tag)
        return []

    # setup: multiprocessing
    num_processes = _parse_core_spec(processes)
    pool = mp.Pool(processes=num_processes)

    # setup: internal state
    num_total = len(data)
    chunksize = max(1, num_total // 1000)
    time_start = time.time()

    # setup: the update_limit decides when to print progress update messages
    update_limit = _ProgressUpdateLimit(
        time_start=time_start,
        num_total=num_total,
        every_n_seconds=update_every_n_seconds,
        every_n_percent=update_every_n_percent
    )

    # before execution
    println(_start_string(num_total, tag, num_processes))

    # actual execution
    result = []
    meta_args = [(func, d) for d in data]
    for idx, r in enumerate(pool.imap(_meta_func, meta_args, chunksize=chunksize)):
        result.append(r)

        num_processed = idx + 1
        if update_limit.should_print(num_processed):
            println(_progress_string(time_start, num_processed, num_total, tag))

    # after execution
    println(_finish_string(time_start, num_total, tag))

    # clean up! See: https://bugs.python.org/issue34172 - Python
    # does NOT terminate the background pool processes by default even
    # though the documentation claims it does so when being GCed.
    pool.close()

    return result


def _filter(
    func,
    data,
    quiet=False,
    processes="n",
    tag="BusyBee",
    stdout=sys.stdout,
    update_every_n_seconds=5,
    update_every_n_percent=50,
):
    """Applies the given `func` to every item in `data` using up to the number of processes
    specified by `processes`. Returns all `data` items where `func` evaluates `True`.

    Args:
        func: The function that will be applied to the `data` items. It needs to be pickleable and
            therefore it must not be a lambda expression.

        data (list): The data that is processed by `func`. This must provide random access and `len()` support.
                     Ideally this is a simple list.

        For the other arguments see the map(...) function.

    Returns:
        The filterred items following the order of the original list.
    """
    is_included = _map(
        func=func,
        data=data,
        quiet=quiet,
        processes=processes,
        tag=tag,
        stdout=stdout,
        update_every_n_seconds=update_every_n_seconds,
        update_every_n_percent=update_every_n_percent,
    )

    result = [item for idx, item in enumerate(data) if is_included[idx]]
    return result


def _mk_dict(
    func,
    keys,
    quiet=False,
    processes="n",
    tag="BusyBee",
    stdout=sys.stdout,
    update_every_n_seconds=5,
    update_every_n_percent=50,
):
    """Creates a new dictionary with the given `keys` and values as the application of `func`
    to the individual key.

    Args:
        func: The function that will be applied to the `keys` items to derive the respective values.
              It needs to be pickleable and therefore it must not be a lambda expression.

        keys (list): The keys of the new dictionary. This must provide random access and `len()` support.
                     Ideally this is a simple list.

        For the other arguments see the map(...) function.

    Returns:
        The a new dictionary with the given `keys` and values as `func(keys)`.
    """

    # Remove duplicates before processing; convert to list to ensure order
    unique_keys = list(set(keys))

    values = _map(
        func=func,
        data=unique_keys,
        quiet=quiet,
        processes=processes,
        tag=tag,
        stdout=stdout,
        update_every_n_seconds=update_every_n_seconds,
        update_every_n_percent=update_every_n_percent,
    )

    return dict(zip(unique_keys, values))



import multiprocessing as mp
import random
import unittest

from .context import _busybee


class processespecParsingTestSuite(unittest.TestCase):

    def test_core_spec_parser_WHEN_number_THEN_number(self):
        self.assertEqual(1, _busybee._parse_core_spec(1))
        self.assertEqual(1, _busybee._parse_core_spec("1"))
        self.assertEqual(8, _busybee._parse_core_spec(8.8))

    def test_core_spec_parser_WHEN_illegal_number_THEN_throws(self):
        illegal_number = [-1, "0", -0.5, 0]

        for spec in illegal_number:
            with self.assertRaises(ValueError):
                _busybee._parse_core_spec(spec)

    def test_core_spec_parser_WHEN_none_like_value_THEN_throws(self):
        with self.assertRaises(ValueError):
            _busybee._parse_core_spec("")

        with self.assertRaises(ValueError):
            _busybee._parse_core_spec(None)

    def test_core_spec_parser_WHEN_just_n_THEN_processes_returned(self):
        n = 8
        self.assertEqual(
            n,
            _busybee._parse_core_spec("n", core_count=lambda: n))

    def test_core_spec_parser_WHEN_n_formula_THEN_calculated_processes_returned(self):
        for n in (4, 6, 8, 32):
            self.assertEqual(
                n-2,
                _busybee._parse_core_spec("n-2", core_count=lambda: n))
            self.assertEqual(
                n//2,
                _busybee._parse_core_spec("n/2", core_count=lambda: n))

    def test_core_spec_parser_WHEN_floating_n_formula_THEN_rounded_down(self):
        self.assertEqual(
            1,
            _busybee._parse_core_spec("n/2", core_count=lambda: 3))

    def test_core_spec_parser_WHEN_n_formula_goes_too_low_THEN_calculated_processes_always_positive(self):
        n = 8
        self.assertEqual(
            1,
            _busybee._parse_core_spec("n-999", core_count=lambda: n))
        self.assertEqual(
            1,
            _busybee._parse_core_spec("n/999", core_count=lambda: n))

    def test_core_spec_parser_WHEN_illegal_n_formula_THEN_throws(self):
        illegal_statements = [
            "n+1",
            "m-1",
            "1-n",
            "****",
            "n-n",
            "n-"
        ]

        for stmt in illegal_statements:
            with self.assertRaises(ValueError):
                _busybee._parse_core_spec(stmt)


class MetaFuncTestSuite(unittest.TestCase):

    def test_meta_func_WHEN_called_THEN_func_executed_on_data(self):
        func, data = lambda x: x+1, 1
        args = (func, data)

        actual = _busybee._meta_func(args)
        self.assertEqual(2, actual)


class ProgressUpdateLimitTestSuite(unittest.TestCase):

    def test_progress_update_limit_WHEN_time_passes_THEN_issues_correct_updates(self):
        pul = _busybee._ProgressUpdateLimit(
            time_start=0.0,
            num_total=100,
            every_n_seconds=5,
            every_n_percent=None,
        )

        self.assertEqual(False, pul.should_print(0, lambda: 1.0))
        self.assertEqual(True, pul.should_print(0, lambda: 5.0))
        self.assertEqual(False, pul.should_print(0, lambda: 5.0))
        self.assertEqual(False, pul.should_print(0, lambda: 9.9999))
        self.assertEqual(True, pul.should_print(0, lambda: 20))
        self.assertEqual(False, pul.should_print(0, lambda: 20))

    def test_progress_update_limit_WHEN_percent_passes_THEN_issues_correct_updates(self):
        pul = _busybee._ProgressUpdateLimit(
            time_start=0.0,
            num_total=100,
            every_n_seconds=None,
            every_n_percent=25,
        )

        self.assertEqual(False, pul.should_print(0))
        self.assertEqual(False, pul.should_print(24))
        self.assertEqual(True, pul.should_print(25))
        self.assertEqual(False, pul.should_print(25))
        self.assertEqual(True, pul.should_print(80))
        self.assertEqual(False, pul.should_print(80))

    def test_progress_update_limit_WHEN_all_off_THEN_no_updates(self):
        pul = _busybee._ProgressUpdateLimit(
            time_start=0.0,
            num_total=100,
            every_n_seconds=None,
            every_n_percent=None,
        )

        self.assertEqual(False, pul.should_print(0, lambda: 20))
        self.assertEqual(False, pul.should_print(10, lambda: 200))
        self.assertEqual(False, pul.should_print(100, lambda: 2000))

    def test_progress_update_limit_WHEN_100_percent_THEN_no_update(self):
        pul = _busybee._ProgressUpdateLimit(
            time_start=0.0,
            num_total=100,
            every_n_seconds=1,
            every_n_percent=1,
        )

        self.assertEqual(False, pul.should_print(100, lambda: 100))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="busybee",
    version="0.1.2",
    author="Daniel H",
    author_email="not_provided@example.org",
    description="Simple and interactive multi-processing for Python and notebooks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lambdapioneer/busybee",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
    ],
    python_requires='>=3.6',
)



import multiprocessing as mp
import random
import unittest

from .context import _string_helpers as _sh


class RelativeTimeStringTestSuite(unittest.TestCase):

    def test_rel_time_string_WHEN_zero_THEN_zero_ms(self):
        self.assertEqual("0ms", _sh._relative_time_string(0.0))
        self.assertEqual("0ms", _sh._relative_time_string(-0.0))

    def test_rel_time_string_WHEN_none_THEN_dash(self):
        self.assertEqual("-", _sh._relative_time_string(None))

    def test_rel_time_string_WHEN_in_ms_area_THEN_ms_string(self):
        self.assertEqual("100ms", _sh._relative_time_string(0.1))
        self.assertEqual("999ms", _sh._relative_time_string(0.999))
        self.assertEqual("-50ms", _sh._relative_time_string(-0.05))

    def test_rel_time_string_WHEN_in_seconds_area_THEN_seconds_string(self):
        self.assertEqual("1.0s", _sh._relative_time_string(1.0))
        self.assertEqual("60.2s", _sh._relative_time_string(60.2))
        self.assertEqual("-5.0s", _sh._relative_time_string(-5))

    def test_rel_time_string_WHEN_in_minutes_area_THEN_minutes_string(self):
        self.assertEqual("2:02m", _sh._relative_time_string(120 + 2))
        self.assertEqual("-2:05m", _sh._relative_time_string(-120 - 5))
        self.assertEqual("5:00m", _sh._relative_time_string(5*60))

    def test_rel_time_string_WHEN_in_hours_area_THEN_hours_string(self):
        self.assertEqual("1:00:00h", _sh._relative_time_string(3600))
        self.assertEqual(
            "1:23:45h",
            _sh._relative_time_string(3600 + 23*60 + 45))
        self.assertEqual(
            "-11:23:45h",
            _sh._relative_time_string(-11*3600 - 23*60 - 45))


class StartStringTestSuite(unittest.TestCase):

    def test_start_string_WHEN_given_info_THEN_all_in_output(self):
        actual = _sh._start_string(100, "tag", 8)
        self.assertIn("tag:", actual)
        self.assertIn("100 items", actual)
        self.assertIn("8 processes", actual)


class FinishStringTestSuite(unittest.TestCase):

    def test_finish_string_WHEN_given_info_THEN_all_in_output(self):
        actual = _sh._finish_string(
            time_start=0.0,
            num_total=100,
            tag="tag",
            current_time=lambda: 4.2
        )
        self.assertIn("tag:", actual)
        self.assertIn("100 items", actual)
        self.assertIn("in 4.2s", actual)
        self.assertIn("avg: 42ms", actual)

    def test_finish_string_WHEN_given_zeros_THEN_output_valid(self):
        actual = _sh._finish_string(
            time_start=0.0,
            num_total=0,
            tag="tag",
            current_time=lambda: 0.0
        )
        self.assertIn("tag:", actual)
        self.assertIn("0 items", actual)
        self.assertIn("in 0.0s", actual)
        self.assertIn("avg: -", actual)


class ProgressStringTestSuite(unittest.TestCase):

    def test_progress_string_WHEN_given_info_THEN_all_in_output(self):
        actual = _sh._progress_string(
            time_start=0.0,
            num_processed=42,
            num_total=100,
            tag="tag",
            current_time=lambda: 42
        )
        self.assertIn("tag:", actual)
        self.assertIn("42/100", actual)
        self.assertIn("42.0%", actual)
        self.assertIn("avg: 1.0s", actual)
        self.assertIn("rem: 58.0s", actual)

    def test_progress_string_WHEN_given_zeros_THEN_output_valid(self):
        actual = _sh._progress_string(
            time_start=0.0,
            num_processed=0,
            num_total=0,
            tag="tag",
            current_time=lambda: 0
        )
        self.assertIn("tag:", actual)
        self.assertIn("0/0", actual)
        self.assertIn("0%", actual)
        self.assertIn("avg: -", actual)
        self.assertIn("rem: -", actual)
